/**
 ** forceFunction.c
 **
 ** Functions for codels of module force
 **
 ** Author: Daniel Sidobre
 ** Date: 20 juillet 2006
 **
 **/
/*
 *  28 novembre 2007 : ghange fft lib
 *     Now use kiss_fft  from http://sourceforge.net/projects/kissfft/
 *     BSD type licence and more possibilities
 */

/* #include <portLib.h> */
#include "server/forceHeader.h"

#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "jr3DriverInterface.h"
#include "forceFunctionHeader.h"
#include "forceConst.h"
#include <sys/time.h>

#include "kiss_fftr.h"

void forceUnitsInit()
{
  /* Choice of filter for permanent conversion */ 
  //SDI_F->filter = 0; /* 8kHz no filter */
  //SDI_F->filter = 1; /* 500 Hz */
  SDI_F->filter = 2; /* 125 Hz */
  //SDI_F->filter = 3; /* 31.5 */
  //SDI_F->filter = 4; /* 7.8125 */
  //SDI_F->filter = 5; /* 1.95 */
  //SDI_F->filter = 6; /* 0.488 */
  SDI_F->fullScale.x  = 1;
  SDI_F->fullScale.y  = 1;
  SDI_F->fullScale.z  = 1;
  SDI_F->fullScale.rx = 1;
  SDI_F->fullScale.ry = 1;
  SDI_F->fullScale.rz = 1;
  SDI_F->offset.x  = 0;
  SDI_F->offset.y  = 0;
  SDI_F->offset.z  = 0;
  SDI_F->offset.rx = 0;
  SDI_F->offset.ry = 0;
  SDI_F->offset.rz = 0;
}

void forceJr3ToNewton(Gb_force* fullScale, Gb_force* offset,
		      jr3DSPStruct_f8* sensorForce, Gb_force* force)
{
  force->x  = ((double) sensorForce->fx) * fullScale->x  + offset->x ;
  force->y  = ((double) sensorForce->fy) * fullScale->y  + offset->y ;
  force->z  = ((double) sensorForce->fz) * fullScale->z  + offset->z ;
  force->rx = ((double) sensorForce->mx) * fullScale->rx + offset->rx;
  force->ry = ((double) sensorForce->my) * fullScale->ry + offset->ry;
  force->rz = ((double) sensorForce->mz) * fullScale->rz + offset->rz;
}
  

void recordWrite(jr3DSPStruct_f8 recordData[], int nbData,
		 Gb_force* fullScale, Gb_force* offset, double dectectionValue[], double detectThreshold, char* fileName)
{
  Gb_force force;
  int i;
  FILE *file = 0;
  static int number = 0;
  char fileNameNumber[512];

  sprintf(fileNameNumber, "%s%d", fileName, number++);
  file = fopen(fileNameNumber, "w");
  if (file == 0) {
    perror("recordWrite fopen Pb:");
    fprintf(stderr, "%s:%d: Pb open filename =\"%s\"\n",
	    __FILE__, __LINE__, fileName);
    return;
  }
  fprintf(file, "# i fx fy fz mx my mz detect_Value detect_Threshold ;\n");
  for (i = 0; i < nbData; i++) {
    forceJr3ToNewton(fullScale, offset, &(recordData[i]), &force);
    fprintf(file, "%d %g %g %g %g %g %g %g %g ;\n", i, force.x, force.y, force.z,
	    force.rx, force.ry, force.rz, dectectionValue[i],detectThreshold);
  }
  fclose(file);
}



/*
 *    FFT
 *
 */

static int forceFft_nb = 0;
static kiss_fft_scalar* forceFft_input = 0;
static kiss_fft_scalar* forceFft_output_i = 0;
static kiss_fft_scalar* forceFft_outputP = 0;
static kiss_fft_cpx* forceFft_output = 0;
static kiss_fftr_cfg forceFftr_cfg = 0;  /* FFT real -> complex direct */
static kiss_fftr_cfg forceFftri_cfg = 0; /* FFT real -> complex inverse */
static double *forceFft_time = 0;

#define ERROR_RETURN(MESS, RETURNVALUE)					\
  do {									\
    fprintf(stderr, "%s:%d: Pb " #MESS "\n", __FILE__, __LINE__);	\
    return RETURNVALUE;							\
  } while (0)

int forceFftIsInit()
{
  printf("forceFft_nb %d\n", forceFft_nb);
  return forceFft_nb;
}

int forceFftInit(nb)
{
  if (nb == forceFft_nb) return OK;
  forceFft_nb = 0;
  if (forceFft_input != 0) free(forceFft_input);
  forceFft_input = (double*) malloc(nb * sizeof(double));
  if (forceFft_input == 0) ERROR_RETURN("malloc", ERROR);

  if (forceFft_output != 0) free(forceFft_output);
  forceFft_output = (kiss_fft_cpx*) malloc(nb * sizeof(kiss_fft_cpx));
  if (forceFft_output == 0) ERROR_RETURN("malloc", ERROR);

  if (forceFft_output_i != 0) free(forceFft_output_i);
  forceFft_output_i = (double*) malloc(nb * sizeof(double));
  if (forceFft_output_i == 0) ERROR_RETURN("malloc", ERROR);

  if (forceFftr_cfg != 0) free(forceFftr_cfg);
  forceFftr_cfg = kiss_fftr_alloc(nb, 0 /*direct*/, 0, 0);
  if (forceFftr_cfg == 0) ERROR_RETURN("malloc", ERROR);

  if (forceFftri_cfg != 0) free(forceFftri_cfg);
  forceFftri_cfg = kiss_fftr_alloc(nb, 1 /*inverse*/, 0, 0);
  if (forceFftri_cfg == 0) ERROR_RETURN("malloc", ERROR);

  if (forceFft_outputP != 0) free(forceFft_outputP);
  forceFft_outputP = (double*) malloc((nb/2+1) * sizeof(double));
  if (forceFft_outputP == 0) ERROR_RETURN("malloc", ERROR);

  if (forceFft_time != 0) free(forceFft_time);
  forceFft_time = malloc(nb * sizeof(double));
  if (forceFft_time == 0) ERROR_RETURN("malloc", ERROR);
  forceFft_nb = nb;
  printf("forceFftInit: forceFft_nb= %d\n", forceFft_nb);
  return OK;
}

void force_kiss_fftr(kiss_fftr_cfg st, kiss_fft_scalar *timedata,
		     kiss_fft_cpx *freqdata);


/*
 *   forceFftP : Fast Fourier Transform : outputP = sqrt(s^2 + c^2)
 *
 */
int forceFftP()
{
  double error;
  double error1;
  int i;
  double s;
  double c;

  //  force_kiss_fftr(forceFftr_cfg, forceFft_input, forceFft_output);
  kiss_fftr(forceFftr_cfg, forceFft_input, forceFft_output);
  forceFft_outputP[0] = 0.; // easier to interpret... Be careful
  for (i = 1; i < (forceFft_nb / 2 +1); i++) {
    c = forceFft_output[i].r;
    s = forceFft_output[i].i;
    forceFft_outputP[i] = sqrt(c*c + s*s);
  }
  {
    /* DEBUG */
    //    force_kiss_fftr(forceFftri_cfg, forceFft_output_i, forceFft_output);
    kiss_fftri(forceFftri_cfg, forceFft_output, forceFft_output_i);
    error = 0.0;
    for (i = 0; i < forceFft_nb; i++) {
      error1 = 
	fabs(forceFft_input[i] - (forceFft_output_i[i] /forceFft_nb) );
      if (error1 > error) 
	error = error1;
    }
    if (error >= 1e-7) {
      for (i = 0; i< forceFft_nb; i++)
	printf(" %g  \t%g\n", forceFft_input[i], forceFft_output_i[i]);
      fprintf(stderr, "error= %g  >= 1e-7\n", error);
      // ERROR_RETURN("error >= 1e-7", ERROR);
    }
  }
  return OK;
}

#define DEBUG_FORCE 1
void forceFftPush(double force)
{
  double* pForce;
  double* pMaxForce; // pointer to last element of forceFft_input
  int size = forceFft_nb;

  if (forceFft_input == 0) return;
  pMaxForce = &(forceFft_input[size-1]);
  for (pForce = forceFft_input; pForce < pMaxForce; pForce++) {
    pForce[0] = pForce[1];
  }
  *pMaxForce = force;    
  
#ifdef DEBUG_FORCE
  {  // to debug time problems
    struct timeval tt;
    struct timezone tz;
    static int premier = 1;
    static struct timeval t0;
    
    pMaxForce = &(forceFft_time[size-1]);
    for (pForce = forceFft_time; pForce < pMaxForce; pForce++) {
      pForce[0] = pForce[1];
    }
    tz.tz_minuteswest = 0;
    tz.tz_dsttime = 0;
    gettimeofday(&tt, &tz);
    if (premier) {
      premier = 0;
      t0 = tt;
    }
    *pMaxForce = tt.tv_sec - t0.tv_sec + (tt.tv_usec - t0.tv_usec) * 0.000001;
  }
#endif
}


int RecordforceFftBlt(const char* fileName)
{
  int i;
  FILE *file;
  double fft;
  char fileNameNumber[512];
  int static number = 0;
  double time;
  
  forceFftP();
  sprintf(fileNameNumber, "%s%d", fileName, number++);
  file = fopen(fileNameNumber, "w");
  if (file == 0) return ERROR;
  fprintf(file, "# i force fft t fHz ;\n");
  // fin pas beau
  for (i = 0; i < forceFft_nb; i++) {
    if (i < forceFft_nb / 2 +1) {
      fft = forceFft_outputP[i];
    } else {
      fft = 0.;
    }
#ifdef DEBUG_FORCE
    time = forceFft_time[i];
#else
    time = -1.;
#endif
    fprintf(file, "%d %g %g %g %g ;\n",
	    i, forceFft_input[i], fft, time,
	    (i *FORCE_NB_TIC_SEC/(FORCE_PERIOD*1.))/forceFft_nb); // 100= nb_tic/s
  }
  fclose(file);
  return OK;
}

double forceFftMeanFirst(int nb)
{
  int i;
  double ret = 0.;

  for (i = 1; i <= nb; i++) {
    ret += forceFft_outputP[i];
  }
  return ret;
}


double forceFftDirection(Gb_force* force, Gb_force* fftDirection)
{
  double f;
  f = 0.0;
  f += force->x  * fftDirection->x ;
  f += force->y  * fftDirection->y ;
  f += force->z  * fftDirection->z ;
  f += force->rx * fftDirection->rx;
  f += force->ry * fftDirection->ry;
  f += force->rz * fftDirection->rz;
  return f;
}
