/**
 ** forceMainTaskCodels.c
 **
 ** Codels called by execution task forceMainTask
 **
 ** Author: Daniel Sidobre
 ** Date: 20 juillet 2006
 **
 **/

#ifdef VXWORKS
# include <vxWorks.h>
#else
# include <portLib.h>
#endif
#include "server/forceHeader.h"

#include "jr3DriverInterface.h"
#include "forceFunctionHeader.h"
#include <math.h>

/* For reading sensor datas */
static jr3DSPStruct sensorData;

/* For records forces */
#define RECORD_SIZE 1000
jr3DSPStruct_f8 recordData[RECORD_SIZE];
double recordDetectionValue[RECORD_SIZE];

/*------------------------------------------------------------------------
 *
 * forceMainTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
forceMainTaskInit(int *report)
{
  /* do nothing : Init transfered in Init request */
  SDI_F->state.moduleInitialized = 0;
  SDI_F->record_i = -1;

  return OK;
}

/*------------------------------------------------------------------------
 *
 * forceMainTaskEnd  --  Termination codel
 *
 * Description: 
 * 
 * Returns:    OK or ERROR
 */

STATUS
forceMainTaskEnd(void)
{
  SDI_F->state.moduleInitialized = 0;
  if (jr3Close() != 0) {
    return ERROR;
  }
  return OK;
}

/*------------------------------------------------------------------------
 *
 * forceMainTaskPerm  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: 
 * 
 * Reports:      OK
 *              S_force_FORCE_ERROR
 *              S_force_SensorError
 * 
 * Returns:    OK or ERROR
 */

STATUS
forceMainTaskPerm(int *report)
{
  jr3DSPStruct_f8* sensorForce;

  /* we do nothing until init */
  if (SDI_F->state.moduleInitialized == 0) {
    return OK;
  }
  if (jr3DataRead(&sensorData) != sizeof(sensorData)) {
    fprintf(stderr, "forceMainTaskPerm: jr3DataRead error\n");
    *report = S_force_SensorError;
    return ERROR;
  }
  switch (SDI_F->filter) {
  case 0: sensorForce = &(sensorData.f0); break; // 8000
  case 1: sensorForce = &(sensorData.f1); break; //  500
  case 2: sensorForce = &(sensorData.f2); break; //  125
  case 3: sensorForce = &(sensorData.f3); break; //   32
  case 4: sensorForce = &(sensorData.f4); break; //    8
  case 5: sensorForce = &(sensorData.f5); break; //    2
  case 6: sensorForce = &(sensorData.f6); break; //    0.5
  default:
    fprintf(stderr, "forceMainTaskPerm: switch pb");
    *report = S_force_FORCE_ERROR;
    return ERROR;
  }
  forceJr3ToNewton(&(SDI_F->fullScale), &(SDI_F->offset),
		   sensorForce, &(SDI_F->force));
  forceFftPush(forceFftDirection(&(SDI_F->force), &(SDI_F->fftDirection)));
  if ( SDI_F->record_i >= 0) {
    recordData[SDI_F->record_i] = *sensorForce;
    recordDetectionValue[SDI_F->record_i] = SDI_F->detection.value ;
    (SDI_F->record_i)++;
    if (SDI_F->record_i >= RECORD_SIZE) {
      recordWrite(recordData, RECORD_SIZE, &(SDI_F->fullScale),
		  &(SDI_F->offset), recordDetectionValue,
		  SDI_F->detection.threshold,
		  "/usr/local/openrobots/data/force/forcesRecord.data");
      SDI_F->record_i = -1;
    }
  }
  return OK;
}
/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_force_Jr3InitFailed
 *              S_force_FFTInitFailed
 */

/* forceInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
forceInitMain(int *report)
{
  //  SDI_F->state.numberFFT = 1 << 10; /* 1024 must be a power of two */
  SDI_F->state.numberFFT = 1 << 9; /* 512 must be a power of two */
  if (forceFftInit(SDI_F->state.numberFFT) != OK) {
    SDI_F->state.numberFFT = 0;
    *report = S_force_FFTInitFailed;
    return ETHER;
  }
  /* force sensor */
  if (jr3Init() != 1) {
    *report = S_force_Jr3InitFailed;
    return ETHER;
  }
  forceUnitsInit();
  /* direction for FFT */
  SDI_F->fftDirection.x  = 1/sqrt(3);
  SDI_F->fftDirection.y  = 1/sqrt(3);
  SDI_F->fftDirection.z  = 1/sqrt(3);
  SDI_F->fftDirection.rx = 0;
  SDI_F->fftDirection.ry = 0;
  SDI_F->fftDirection.rz = 0;
  //  SDI_F->state.sensor;
  SDI_F->detection.threshold = 6000;
  SDI_F->detection.nb_val_fft = 7;
  SDI_F->state.moduleInitialized = 1;
  return ETHER;
}

/*------------------------------------------------------------------------
 * Detect
 *
 * Description: 
 *
 * Reports:      OK
 *              S_force_NotInitialized
 */

/* forceDetectStart  -  codel START of Detect
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
forceDetectStart(int *report)
{
  if (SDI_F->state.moduleInitialized == 0) {
    *report = S_force_NotInitialized;
    return ETHER;
  }
  if (forceFftIsInit() <= 0) return ETHER;
  forceFftP();
  //  printf("forceDetectStart  forceFftP 1\n");
  forceFftP();
  // printf("forceDetectStart  forceFftP 2\n");
  SDI_F->detection.value = forceFftMeanFirst(SDI_F->detection.nb_val_fft);
  if (SDI_F->detection.value > SDI_F->detection.threshold) {
    SDI_F->detection.grasped = 2;
    /*    *report = S_force_InitialStateIncompatible;
     *	  return ETHER;
     */
  } else {
    SDI_F->detection.grasped = 0;
  }
  printf("forceDetectStart  start OK\n");
  return EXEC;
}

/* forceDetectMain  -  codel EXEC of Detect
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
forceDetectMain(int *report)
{
  forceFftP();
  // printf("forceDetectMain forceFftP ok\n");
  SDI_F->detection.value = forceFftMeanFirst(SDI_F->detection.nb_val_fft);
  //  printf("forceDetectMain forceFftMeanFirst ok\n");
  if (SDI_F->detection.value > SDI_F->detection.threshold) {
    if (SDI_F->detection.grasped == 2) {
      /* we wait detection.value < detection.threshold as initial state */
      return EXEC;
    }
    SDI_F->detection.grasped = 1;
    printf("forceDetectMain     SDI_F->detection.grasped = 1;\n");
    return ETHER;
  } else {
    SDI_F->detection.grasped = 0;
  }
  return EXEC;
}

/*------------------------------------------------------------------------
 * MonitorForces
 *
 * Description: 
 *
 * Reports:      OK
 *              S_force_NotInitialized
 */

/* forceMonitorForcesStart  -  codel START of MonitorForces
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
forceMonitorForcesStart(FORCE_MONITOR_DATA_STR *eq, int *report)
{
  jr3DSPStruct_f8* sensorForce = &(sensorData.f6);

  if (SDI_F->state.moduleInitialized == 0) {
    *report = S_force_NotInitialized;
    return ETHER;
  }
  forceJr3ToNewton(&(SDI_F->fullScale), &(SDI_F->offset),
		   sensorForce, &(SDI_F->monitor.force0));
  return EXEC;
}

/* forceMonitorForcesMain  -  codel EXEC of MonitorForces
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
forceMonitorForcesMain(FORCE_MONITOR_DATA_STR *eq, int *report)
{
  FORCE_MONITOR_STR* monitor = &(SDI_F->monitor);

  monitor->value = 0.;
  monitor->value += (SDI_F->force.x  - monitor->force0.x)  * monitor->eq.direction.x;
  monitor->value += (SDI_F->force.y  - monitor->force0.y)  * monitor->eq.direction.y;
  monitor->value += (SDI_F->force.z  - monitor->force0.z)  * monitor->eq.direction.z;
  monitor->value += (SDI_F->force.rx - monitor->force0.rx) * monitor->eq.direction.rx;
  monitor->value += (SDI_F->force.ry - monitor->force0.ry) * monitor->eq.direction.ry;
  monitor->value += (SDI_F->force.rz - monitor->force0.rz) * monitor->eq.direction.rz;
  if (monitor->value >= eq->threshold) {
    return ETHER;
  }
  return EXEC;
}

/*------------------------------------------------------------------------
 * SaveFileFFT
 *
 * Description: 
 *
 * Reports:      OK
 *              S_force_CANNOT_OPEN_FILE
 */

/* forceSaveFileFFTMain  -  codel EXEC of SaveFileFFT
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
forceSaveFileFFTMain(int *report)
{
  if (RecordforceFftBlt("/usr/local/openrobots/data/force/forceFFT.dat")
      == ERROR) {
    *report = S_force_CANNOT_OPEN_FILE;
  }
  return ETHER;
}

void forceOffsetReset()
{
  SDI_F->offset.x  = - sensorData.f6.fx;
  SDI_F->offset.y  = - sensorData.f6.fy;
  SDI_F->offset.z  = - sensorData.f6.fz;
  SDI_F->offset.rx = - sensorData.f6.mx;
  SDI_F->offset.ry = - sensorData.f6.my;
  SDI_F->offset.rz = - sensorData.f6.mz;
}




